"""
Constants used throughout Lumos
"""

SPEED_OF_LIGHT = 3.0 * 1e8 # meters / second
WAVELENGTH = 532 * 1e-9     # meters
EARTH_RADIUS = 6378 * 1000  # meters
SUN_INTENSITY = 1360 # Watts / meter ^ 2